unit DCTII;

interface
uses Transformation;

function DCTInit(X,Y:longword):TTransError; cdecl;
function DCTFinalize : TTransError; cdecl;
procedure DCTGetEM(Mask:pointer; Energy:pointer) cdecl;
function DCTGetBVal(FnX,FnY,X,Y:longword):TFloat; cdecl;
function DCTTransform(Dir:TTransDirection; Data:pointer):TTransError; cdecl;
function DCTGetP(ID:longword;var Value:pointer; var Len:longword):TTransError; cdecl;
function DCTSetP(ID:longword;Value:pointer;Len:longword):TTransError; cdecl;
function DCTGetConf(var Value:pointer):longword; cdecl;
procedure DCTSetConf(Value:pointer;Len:longword); cdecl;
function DCTGetQuantM(X,Y,Width,Height:longword):TFloat; cdecl;
function DCTGetStatus : TTransStatus; cdecl;


var
  DCT : TTransformation = (
    Name : 'DCT II';
    Description : 'Discrete Cosine Transformation II';
    ID : 2;
    Version : $100;
    Public : true;
    Init : DCTInit;
    Finalize : DCTFinalize;
    GetEnergyMatrix : DCTGetEM;
    GetBasisValue : DCTGetBVal;
    Transform : DCTTransform;
    GetProperty : DCTGetP;
    SetProperty : DCTSetP;
    GetConfig : DCTGetConf;
    SetConfig : DCTSetConf;
    Configure : nil;
    GetQuantMatrixItem : DCTGetQuantM;
    Status : DCTGetStatus;
  );


implementation
uses DOT;

// DCT II Normalization matrix for pictures of size 8x8, JPEG standard
const DCTIINormMatrix : array[0..7,0..7] of byte =
      ((16,11,10,16,24,40,51,61),
       (12,12,14,19,26,58,60,55),
       (14,13,16,24,40,57,69,56),
       (14,17,22,29,51,87,80,62),
       (18,22,37,56,68,109,103,77),
       (24,35,55,64,81,104,113,92),
       (49,64,78,87,103,121,120,101),
       (72,92,95,98,112,100,103,99));



var
  Basis1,Basis2 : TShapeData;


function DCTInit;
var i:integer;
begin
  //Initialize internal transformation to DCTII
  DCT.LastError:=TTransError(DOT.Initialize(X,Y,dmDCTII));
  if DCT.LastError=teOK then begin
    DCT.X:=X;
    DCT.Y:=Y;
    try
      SetLength(Basis1,4*X);
      SetLength(Basis2,4*Y);
      //Make Basis generator arrays
      for i:=0 to 4*X-1 do Basis1[i]:=cos(i*Pi/(2*X));
      for i:=0 to 4*Y-1 do Basis2[i]:=cos(i*Pi/(2*Y));
    except
      DOT.Finalize;
      DCT.LastError:=teNotEnoughMem;
    end;
  end;
  Result:=DCT.LastError;
end;


function DCTFinalize;
begin
  if DOT.GetMode<>dmDCTII then begin DCT.LastError:=teIncorrectMode; Result:=teIncorrectMode; exit end;
  DCT.LastError:=TTransError(DOT.Finalize);
  SetLength(Basis1,0); Basis1:=nil;
  SetLength(Basis2,0); Basis2:=nil;
  Result:=DCT.LastError;
end;


procedure DCTGetEM;
var
    iTemp, iEnergy : TShapeData;
    u,v,k,l:integer;
    ck,cl,cu,cv:TFloat;
    alpha,beta : shortint;
    sqrt2 : TFloat;
    norma : TFloat;

begin
  if DOT.GetMode<>dmDCTII then begin DCT.LastError:=teIncorrectMode; exit; end;
  iTemp:=Mask; iEnergy:=Energy;
  sqrt2:=sqrt(2); norma:=2*sqrt(DCT.X*DCT.Y);
  DCT.LastError:=DCT.Transform(tdForward,Mask);
  for v:=0 to DCT.Y-1 do begin
    if DCT.Y=2*v then beta:=0 else beta:=(DCT.Y-2*v) div abs(DCT.Y-2*v);
    if beta=1 then l:=2*v else l:=2*(DCT.Y-v);
    if l=0 then cl:=sqrt2 else cl:=1;
    if v=0 then cv:=2 else cv:=1;
    for u:=0 to DCT.X-1 do begin
      if DCT.X=2*u then alpha:=0 else alpha:=(DCT.X-2*u) div abs(DCT.X-2*u);
      if alpha=1 then k:=2*u else k:=2*(DCT.X-u);
      if k=0 then ck:=sqrt2 else ck:=1;
      if u=0 then cu:=2 else cu:=1;
      iEnergy[v*DCT.X+u]:=(iTemp[0]*2 + alpha*iTemp[k]*(ck*sqrt2) + beta*iTemp[l*DCT.X]*(cl*sqrt2)
                           + alpha*beta*iTemp[l*DCT.X+k]*(ck*cl))/(norma*cu*cv);
    end;
  end;
end;


function DCTGetBVal;
begin
  if DOT.GetMode<>dmDCTII then begin DCT.LastError:=teIncorrectMode; Result:=0; exit end;
  if DOT.State=dsClosed then begin DCT.LastError:=teClosed; Result:=0; exit end;
  Result:=Basis1[((2*X+1)*FnX) mod (DCT.X shl 2)]
         *Basis2[((2*Y+1)*FnY) mod (DCT.Y shl 2)];
  if FnX=0 then Result:=Result/sqrt(DCT.X) else Result:=Result*sqrt(2/DCT.X);
  if FnY=0 then Result:=Result/sqrt(DCT.Y) else Result:=Result*sqrt(2/DCT.Y);
end;


function DCTTransform;
begin
  if DOT.GetMode<>dmDCTII then begin DCT.LastError:=teClosed; Result:=teIncorrectMode; exit end;
  if DOT.State=dsClosed then begin DCT.LastError:=teClosed; Result:=teClosed; exit end;
  DCT.LastError:=TTransError(DOT.Transform(TDOTDirection(Dir),nil,Data));
  Result:=DCT.LastError;
end;


function DCTGetP;
begin
  Value:=nil;
  Len:=0;
  Result:=tePropertyNotExists;
end;


function DCTSetP;
begin
  Result:=tePropertyNotExists;
end;

function DCTGetConf;
begin
  Value:=nil;
  Result:=0;
end;


procedure DCTSetConf;
begin

end;


function DCTGetQuantM;
var A1,A2 : integer;
    V1, V2 : TFloat;
begin
  A1:=(X shl 3) div Width;
  A2:=(Y shl 3) div Height;
  if Width>7 then begin
    if A1<7 then
      V1:=DCTIINormMatrix[A2][A1]+(8.0*(DCTIINormMatrix[A2][A1+1]-DCTIINormMatrix[A2][A1])*(X mod (Width shr 3)))/Width
    else
      V1:=DCTIINormMatrix[A2][A1];
    if A2<7 then begin
      if A1<7 then
        V2:=DCTIINormMatrix[A2+1][A1]+8.0*(DCTIINormMatrix[A2+1][A1+1]-DCTIINormMatrix[A2+1][A1])*(X mod (Width shr 3))/Width
      else
        V2:=DCTIINormMatrix[A2+1][A1];
    end else
      V2:=V1;
  end else begin
    V1:=DCTIINormMatrix[A2][A1];
    if A2<7 then
      V2:=DCTIINormMatrix[A2+1][A1]
    else
      V2:=V1;
  end;

  if Height>7 then
    Result:=(V1+8*(V2-V1)*(Y mod (Height shr 3))/Height)
  else Result:=V1;
end;


function DCTGetStatus : TTransStatus;
begin
  Result:=TTransStatus(DOT.State);
end;


end.
